@@ -11,6 +11,7 @@ import ctypes
 import ctypes.util
 import socket
 import msgpack
+import os
 
 try:
     from ctypes import c_ssize_t
@@ -21,7 +22,7 @@ import tarantool
 from tarantool.response import Response
 from tarantool.request import (
     Request,
-    RequestOK,
+    # RequestOK,
     RequestCall,
     RequestDelete,
     RequestEval,
@@ -33,28 +34,29 @@ from tarantool.request import (
     RequestSubscribe,
     RequestUpdate,
     RequestUpsert,
-    RequestAuthenticate)
-
+    RequestAuthenticate
+)
 from tarantool.space import Space
 from tarantool.const import (
     SOCKET_TIMEOUT,
     RECONNECT_MAX_ATTEMPTS,
     RECONNECT_DELAY,
-    RETRY_MAX_ATTEMPTS,
     REQUEST_TYPE_OK,
     REQUEST_TYPE_ERROR,
     IPROTO_GREETING_SIZE,
-    ENCODING_DEFAULT)
-
+    ENCODING_DEFAULT,
+    ITERATOR_EQ,
+    ITERATOR_ALL
+)
 from tarantool.error import (
     NetworkError,
-    DatabaseError,
-    warn,
-    RetryWarning,
-    NetworkWarning)
-
-from .schema import Schema
-from .utils import check_key, greeting_decode, version_id
+    # DatabaseError,
+    NetworkWarning,
+    SchemaReloadException,
+    warn
+)
+from tarantool.schema import Schema
+from tarantool.utils import check_key, greeting_decode, version_id
 
 
 class Connection(object):
@@ -89,7 +91,12 @@ class Connection(object):
         creates network connection.
                              if False than you have to call connect() manualy.
         '''
-        libc = ctypes.CDLL(ctypes.util.find_library('c'), use_errno=True)
+        if os.name == 'nt':
+            libc = ctypes.windll.LoadLibrary(
+                ctypes.util.find_library('Ws2_32')
+            )
+        else:
+            libc = ctypes.CDLL(ctypes.util.find_library('c'), use_errno=True)
         recv = self._sys_recv = libc.recv
         recv.argtypes = [
             ctypes.c_int, ctypes.c_void_p, c_ssize_t, ctypes.c_int]
@@ -102,6 +109,7 @@ class Connection(object):
         self.reconnect_delay = reconnect_delay
         self.reconnect_max_attempts = reconnect_max_attempts
         self.schema = Schema(self)
+        self.schema_version = 1
         self._socket = None
         self.connected = False
         self.error = True
@@ -160,6 +168,7 @@ class Connection(object):
             # the connection fails because the server is simply
             # not bound to port
             self._socket.settimeout(self.socket_timeout)
+            self.load_schema()
         except socket.error as e:
             self.connected = False
             raise NetworkError(e)
@@ -169,13 +178,26 @@ class Connection(object):
         while to_read > 0:
             try:
                 tmp = self._socket.recv(to_read)
+            except OverflowError:
+                self._socket.close()
+                err = socket.error(
+                    errno.ECONNRESET,
+                    "Too big packet. Closing connection to server"
+                )
+                raise NetworkError(err)
             except socket.error:
-                raise NetworkError(socket.error(errno.ECONNRESET,
-                                   "Lost connection to server during query"))
+                err = socket.error(
+                    errno.ECONNRESET,
+                    "Lost connection to server during query"
+                )
+                raise NetworkError(err)
             else:
                 if len(tmp) == 0:
-                    raise NetworkError(socket.error(errno.ECONNRESET,
-                                       "Lost connection to server during query"))
+                    err = socket.error(
+                        errno.ECONNRESET,
+                        "Lost connection to server during query"
+                    )
+                    raise NetworkError(err)
                 to_read -= len(tmp)
                 buf += tmp
         return buf
@@ -200,18 +222,17 @@ class Connection(object):
         '''
         assert isinstance(request, Request)
 
-        # Repeat request in a loop if the server returns completion_status == 1
-        # (try again)
-        for attempt in range(RETRY_MAX_ATTEMPTS):    # pylint: disable=W0612
-            self._socket.sendall(bytes(request))
-            response = Response(self, self._read_response())
+        response = None
+        while True:
+            try:
+                self._socket.sendall(bytes(request))
+                response = Response(self, self._read_response())
+                break
+            except SchemaReloadException as e:
+                self.update_schema(e.schema_version)
+                continue
 
-            if response.completion_status != 1:
-                return response
-            warn(response.return_message, RetryWarning)
-
-        # Raise an error if the maximum number of attempts have been made
-        raise DatabaseError(response.return_code, response.return_message)
+        return response
 
     def _opt_reconnect(self):
         '''
@@ -229,8 +250,13 @@ class Connection(object):
                 if e.errno == errno.EBADF:
                     return errno.ECONNRESET
             else:
-                self._sys_recv(sock_fd, buf, 1,
-                               socket.MSG_DONTWAIT | socket.MSG_PEEK)
+                if os.name == 'nt':
+                    flag = socket.MSG_PEEK
+                    self._socket.setblocking(False)
+                else:
+                    flag = socket.MSG_DONTWAIT | socket.MSG_PEEK
+                self._sys_recv(sock_fd, buf, 1, flag)
+
                 if ctypes.get_errno() == errno.EAGAIN:
                     ctypes.set_errno(0)
                     return errno.EAGAIN
@@ -247,7 +273,7 @@ class Connection(object):
             time.sleep(self.reconnect_delay)
             try:
                 self.connect_basic()
-            except NetworkError as e:
+            except NetworkError:
                 pass
             else:
                 if self.connected:
@@ -279,13 +305,20 @@ class Connection(object):
         assert isinstance(request, Request)
 
         self._opt_reconnect()
-        response = self._send_request_wo_reconnect(
-            request)
 
-        return response
+        return self._send_request_wo_reconnect(request)
+
+    def load_schema(self):
+        self.schema.fetch_space_all()
+        self.schema.fetch_index_all()
+
+    def update_schema(self, schema_version):
+        self.schema_version = schema_version
+        self.flush_schema()
 
     def flush_schema(self):
         self.schema.flush()
+        self.load_schema()
 
     def call(self, func_name, *args):
         '''
@@ -363,37 +396,54 @@ class Connection(object):
 
         request = RequestAuthenticate(self, self._salt, self.user,
                                       self.password)
-        return self._send_request_wo_reconnect(request)
+        auth_response = self._send_request_wo_reconnect(request)
+        if auth_response.return_code == 0:
+            self.flush_schema()
+        return auth_response
 
-    def join(self, server_uuid):
+    def _join_v16(self, server_uuid):
         request = RequestJoin(self, server_uuid)
-        sync = request._sync
-        resp = self._send_request(request)
+        self._socket.sendall(bytes(request))
+
         while True:
-            if self.version_id >= version_id(1, 7, 0):
-                # Send acknowledgement
-                ack = RequestOK(self, sync)
-                self._socket.sendall(bytes(ack))
+            resp = Response(self, self._read_response())
             yield resp
             if resp.code == REQUEST_TYPE_OK or resp.code >= REQUEST_TYPE_ERROR:
                 return
-            resp = Response(self, self._read_response())
         self.close()  # close connection after JOIN
 
-    def subscribe(self, cluster_uuid, server_uuid, vclock={}):
-        # FIXME rudnyh: ^ 'vclock={}'? really? sure?
-        request = RequestSubscribe(self, cluster_uuid, server_uuid, vclock)
-        sync = request._sync
-        resp = self._send_request(request)
+    def _join_v17(self, server_uuid):
+        class JoinState:
+            Handshake, Initial, Final, Done = range(4)
+
+        request = RequestJoin(self, server_uuid)
+        self._socket.sendall(bytes(request))
+        state = JoinState.Handshake
         while True:
-            if self.version_id >= version_id(1, 7, 0):
-                # Send acknowledgement
-                ack = RequestOK(self, sync)
-                self._socket.sendall(bytes(ack))
+            resp = Response(self, self._read_response())
             yield resp
             if resp.code >= REQUEST_TYPE_ERROR:
                 return
+            elif resp.code == REQUEST_TYPE_OK:
+                state = state + 1
+                if state == JoinState.Done:
+                    return
+
+    def join(self, server_uuid):
+        self._opt_reconnect()
+        if self.version_id < version_id(1, 7, 0):
+            return self._join_v16(server_uuid)
+        return self._join_v17(server_uuid)
+
+    def subscribe(self, cluster_uuid, server_uuid, vclock=None):
+        vclock = vclock or {}
+        request = RequestSubscribe(self, cluster_uuid, server_uuid, vclock)
+        self._socket.sendall(bytes(request))
+        while True:
             resp = Response(self, self._read_response())
+            yield resp
+            if resp.code >= REQUEST_TYPE_ERROR:
+                return
         self.close()  # close connection after SUBSCRIBE
 
     def insert(self, space_name, values):
@@ -505,7 +555,8 @@ class Connection(object):
             space_name = self.schema.get_space(space_name).sid
         if isinstance(index_name, six.string_types):
             index_name = self.schema.get_index(space_name, index_name).iid
-        request = RequestUpsert(self, space_name, index_name, tuple_value, op_list)
+        request = RequestUpsert(self, space_name, index_name, tuple_value,
+                                op_list)
         return self._send_request(request)
 
     def update(self, space_name, key, op_list, **kwargs):
@@ -644,7 +695,13 @@ class Connection(object):
         offset = kwargs.get("offset", 0)
         limit = kwargs.get("limit", 0xffffffff)
         index_name = kwargs.get("index", 0)
-        iterator_type = kwargs.get("iterator", 0)
+        iterator_type = kwargs.get("iterator")
+
+        if iterator_type is None:
+            iterator_type = ITERATOR_EQ
+            if key is None or (isinstance(key, (list, tuple)) and
+                               len(key) == 0):
+                iterator_type = ITERATOR_ALL
 
         # Perform smart type checking (scalar / list of scalars / list of
         # tuples)
@@ -674,7 +731,7 @@ class Connection(object):
         return Space(self, space_name)
 
     def generate_sync(self):
-        """\
+        '''
         Need override for async io connection
-        """
+        '''
         return 0
